@@ -25,8 +25,33 @@
 /* include this first, before NO_IMPORT_PYGOBJECT is defined */
 #include <pygobject.h>
 #include <gst/gst.h>
-#include <gmodule.h>
 #include <Python.h>
+/*
+ * We need to call dlopen() directly on macOS to workaround a macOS runtime
+ * linker bug. When there are nested dlopen() calls and the second dlopen() is
+ * called from another library (such as gmodule), @loader_path is resolved as
+ * @executable_path and RPATHs are read from the executable (gst-plugin-scanner)
+ * instead of the library itself (libgstges.dylib). This doesn't happen if the
+ * second dlopen() call is directly in the source code of the library.
+ * Previously seen at:
+ * https://gitlab.freedesktop.org/gstreamer/gstreamer/-/issues/1171#note_2290789
+ */
+#ifdef G_OS_WIN32
+#include <gmodule.h>
+#define gstpy_module_open(fname) g_module_open(fname,0)
+#define gstpy_module_error g_module_error
+#define gstpy_module_symbol(module,name,symbol) g_module_symbol(module,name,symbol)
+#else
+#include <dlfcn.h>
+#define gstpy_module_open(fname) dlopen(fname,RTLD_NOW | RTLD_GLOBAL)
+#define gstpy_module_error dlerror
+static inline gboolean
+gstpy_module_symbol (gpointer handle, const char *name, gpointer * symbol)
+{
+  *symbol = dlsym (handle, name);
+  return *symbol != NULL;
+}
+#endif
 
 void *_PyGstElement_Type;
 
@@ -108,12 +133,11 @@ gst_python_plugin_load_file (GstPlugin * plugin, const char *name)
 }
 
 static gboolean
-gst_python_load_directory (GstPlugin * plugin, gchar * path)
+gst_python_load_directory (GstPlugin * plugin, const gchar * path)
 {
   GDir *dir;
   const gchar *file;
   GError *error = NULL;
-  gboolean ret = TRUE;
 
   dir = g_dir_open (path, 0, &error);
   if (!dir) {
@@ -127,7 +151,7 @@ gst_python_load_directory (GstPlugin * plugin, gchar * path)
     if (g_str_has_suffix (file, ".py")) {
       gsize len = strlen (file) - 3;
       gchar *name = g_strndup (file, len);
-      ret &= gst_python_plugin_load_file (plugin, name);
+      gst_python_plugin_load_file (plugin, name);
       g_free (name);
     }
   }
@@ -146,7 +170,9 @@ gst_python_plugin_load (GstPlugin * plugin)
   /* Mimic the order in which the registry is checked in core */
 
   /* 1. check env_variable GST_PLUGIN_PATH */
-  plugin_path = g_getenv ("GST_PLUGIN_PATH");
+  plugin_path = g_getenv ("GST_PLUGIN_PATH_1_0");
+  if (plugin_path == NULL)
+    plugin_path = g_getenv ("GST_PLUGIN_PATH");
   if (plugin_path) {
     char **list;
     int i;
@@ -164,7 +190,9 @@ gst_python_plugin_load (GstPlugin * plugin)
   }
 
   /* 2. Check for GST_PLUGIN_SYSTEM_PATH */
-  plugin_path = g_getenv ("GST_PLUGIN_SYSTEM_PATH");
+  plugin_path = g_getenv ("GST_PLUGIN_SYSTEM_PATH_1_0");
+  if (plugin_path == NULL)
+    plugin_path = g_getenv ("GST_PLUGIN_SYSTEM_PATH");
   if (plugin_path == NULL) {
     char *home_plugins;
 
@@ -211,8 +239,8 @@ plugin_init (GstPlugin * plugin)
   PyGILState_STATE state = 0;
   PyObject *gi, *require_version, *args, *gst, *dict, *pyplugin;
   gboolean we_initialized = FALSE;
-  GModule *libpython;
   gpointer has_python = NULL;
+  const gchar *override_path;
 
   GST_DEBUG_CATEGORY_INIT (pyplugindebug, "pyplugin", 0,
       "Python plugin loader");
@@ -225,23 +253,19 @@ plugin_init (GstPlugin * plugin)
       GST_PLUGIN_DEPENDENCY_FLAG_NONE);
 
   GST_LOG ("Checking to see if libpython is already loaded");
-  if (g_module_symbol (g_module_open (NULL, G_MODULE_BIND_LOCAL),
+  if (gstpy_module_symbol (gstpy_module_open (NULL),
           "_Py_NoneStruct", &has_python) && has_python) {
     GST_LOG ("libpython is already loaded");
   } else {
-    const gchar *libpython_path =
-        PY_LIB_LOC "/libpython" PYTHON_VERSION PY_ABI_FLAGS "." PY_LIB_SUFFIX;
-    GST_LOG ("loading libpython from '%s'", libpython_path);
-    libpython = g_module_open (libpython_path, 0);
-    if (!libpython) {
-      g_critical ("Couldn't g_module_open libpython. Reason: %s",
-          g_module_error ());
+    GST_LOG ("loading libpython by name: %s", PY_LIB_FNAME);
+    if (!gstpy_module_open (PY_LIB_FNAME)) {
+      GST_ERROR ("Couldn't load libpython. Reason: %s", gstpy_module_error ());
       return FALSE;
     }
   }
 
   if (!Py_IsInitialized ()) {
-    GST_LOG ("python wasn't initialized");
+    GST_LOG ("python wasn't already initialized");
     /* set the correct plugin for registering stuff */
     Py_Initialize ();
     we_initialized = TRUE;
@@ -250,11 +274,40 @@ plugin_init (GstPlugin * plugin)
     state = PyGILState_Ensure ();
   }
 
+  if ((override_path = g_getenv ("GST_OVERRIDE_SRC_PATH"))) {
+    gchar *overrides_setup =
+        g_build_filename (override_path, "..", "..", "testsuite",
+        "overrides_hack.py", NULL);
+    FILE *fd = fopen (overrides_setup, "rb");
+
+    if (!fd || PyRun_SimpleFileExFlags (fd, overrides_setup, 1, 0)) {
+      g_free (overrides_setup);
+
+      return FALSE;
+    } else {
+      g_free (overrides_setup);
+      GST_INFO ("Imported overrides setup");
+    }
+  }
+
   GST_LOG ("Running with python version '%s'", Py_GetVersion ());
 
   GST_LOG ("initializing pygobject");
   if (!pygobject_init (3, 0, 0)) {
-    g_critical ("pygobject initialization failed");
+    if (PyErr_Occurred ()) {
+      PyObject *type, *value, *traceback;
+      PyErr_Fetch (&type, &value, &traceback);
+      PyObject *str = PyObject_Str (value);
+      if (str) {
+        g_critical ("pygobject initialization failed: %s\n",
+            PyUnicode_AsUTF8 (str));
+        Py_DECREF (str);
+      }
+      Py_XDECREF (type);
+      Py_XDECREF (value);
+      Py_XDECREF (traceback);
+      PyErr_Clear ();
+    }
     return FALSE;
   }
 
